import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy.modeling import fitting, Fittable1DModel, Parameter, models
from scipy.optimize import curve_fit
from matplotlib.colors import LogNorm
from mpl_toolkits.mplot3d import Axes3D
from astropy.visualization import ZScaleInterval, ImageNormalize, MinMaxInterval, SqrtStretch
from astropy.wcs import WCS
from astropy import units
from astropy.coordinates import SkyCoord
from regions import PixCoord, CircleSkyRegion
from photutils import centroid_com, centroid_1dg, centroid_2dg, CircularAperture, centroid_sources
from astropy.stats import sigma_clipped_stats, sigma_clip
from photutils import datasets, DAOStarFinder
from astropy.nddata import Cutout2D
from astropy.wcs.utils import pixel_to_skycoord
from photutils import aperture_photometry, SkyCircularAperture, SkyCircularAnnulus, CircularAnnulus
from matplotlib.ticker import MaxNLocator
from multiprocessing import Process, Pool
import os
import glob
import time
import warnings
import subprocess



def zscale(img):
    interval = ZScaleInterval()
    vmin, vmax = interval.get_limits(img)
    norm = ImageNormalize(vmin=vmin, vmax=vmax)
    return norm


def robust_sigma(y, zero='ref'):
    eps = 1.0e-20
    if zero == 'ref':
        y0 = 0
    else:
        y0 = np.median(y)
    mad = np.median(np.abs(y-y0))/0.6745
    if mad < eps:
        mad = np.mean(np.abs(y-y0)/0.80)
    if mad < eps:
        sigma = 0
        return sigma
    
    u = (y-y0)/(6*mad)
    uu = u*u
    q = (uu < 1)
    ql = len(q)
    if ql < 3:
        print ('robust_sigma: This distribution is too weird! returning -1')
        sigma = -1
        return sigma
    
    n = np.sum(y)
    numerator = np.sum((y[q]-y0)**2 * (1-uu[q]**4))
    den1 = np.sum((1-uu[q])*(1-5*uu[q]))
    sigma = n*numerator/(den1*(den1-1))
    if sigma > 0:
        return np.sqrt(sigma)
    else:
        return 0


def mw_main(imagefile, templist, sig=4.0, cores=4):
    start = time.time()
    flist = np.genfromtxt(templist, usecols=0, dtype=str)
    print('Starting main program')
    
    pool = Pool(processes=cores)
    #print('Starting main program')
    results = [pool.apply_async(mask_wht, args=(imagefile, filename, sig)) for filename in flist]
    output = np.array([p.get() for p in results])
    print(time.time()-start, 'seconds')

#@ray.remote
def mask_wht(imagefile, exp_file, sig):
    
    img = fits.open(imagefile)[0].data # median.fits
    wht_file = exp_file.replace('fits', 'weight.fits') # Use the weight files that were previously used
    exp = fits.open(exp_file)
    header = exp[0].header
    exp = exp[0].data
    wht = fits.open(wht_file, mode='update')
    wht_data = wht[0].data
    
    #Get values from each variable of the header file, this is the resampled image that has been WCS calibrated
    x1 = header['COMIN1'] - 1
    y1 = header['COMIN2'] - 1
    x2 = x1 + header['NAXIS1']
    y2 = y1 + header['NAXIS2']
    f = header['FLXSCALE']
    
    #Cut a smaller image with x,y axes as defined that matches the resampled version in shape.
    cut = img[y1:y2,x1:x2] # This is from median.fits
    
    #Find the difference between the cut image and the actual exposure changed to fluxscale
    #Cut has been median combined so there are no outliers.
    diff = cut - exp*f # median cut - resamp exp*f
    g = (wht_data != 0) # We set all the bad pixels to 0 in wht_mask
    
    positive_cut = cut
    positive_cut[cut<0] = 0
    #Find the rms of the median image.
    rms = np.sqrt(positive_cut)
    #Add the noise of the resampled image in quadrature to get an estimate of the noise for a mean image.
    rms = np.sqrt(rms**2 + (robust_sigma(exp[g])*f)**2)
    #Locate all the pixels that contain a signal large than (3) sigma and weight out those values.
    a = (np.abs(diff) > rms*sig)
    print('Processing %s' % exp_file)
    if a.any() == True:
        wht_data[a] = 0
        wht[0].data = wht_data
        #t1 = fits.PrimaryHDU(cut)
        #t1.writeto('cut.fits', overwrite=True)
        ###TRY PRINTINT THESE ONES AND SEE IF THEY ARE THE SAME AS GDL??????
        #t2 = fits.PrimaryHDU(exp)
        #t2.writeto('exp.fits', overwrite=True)
        #t3 = fits.PrimaryHDU(wht_data)
        #t3.writeto('wht.fits', overwrite=True)
        new_wht = fits.PrimaryHDU(wht[0].data)
        new_wht.writeto(wht_file, overwrite=True)
        print(round(np.count_nonzero(a)/(float(x2-x1)*float(y2-y1))*100, 2), '% were masked out')
        print(np.count_nonzero(a), 'pixels were masked out')
    else:
        print("No pixels rejected")


def wm_main(filelist, cores=4):
    start = time.time()
    filelist = np.genfromtxt(filelist, usecols=0, dtype=str)
    pool = Pool(processes=cores)
    print('Starting main program')
    results = [pool.apply_async(wht_mask, args=(i, filename)) for i, filename in enumerate(filelist)]
    output = np.array([p.get() for p in results])
    print(time.time()-start, 'seconds')


def wht_mask(iter,filename):
    start = time.time()
    img=fits.open(filename)
    new_hdu = fits.HDUList()
    new_hdu.append(img[0])
    for i in range(1,len(img)):
        data = img[i].data
        mask = (data<-100)
        wht = data*0+1.0
        wht[mask] = 0
        new_hdu.append(fits.ImageHDU(data=wht))
    #primaryhdu = fits.PrimaryHDU(data=new_hdu, header=img[0].header)
    #primaryhdu.writeto(filename.replace('.fits', '.weight.fits'), overwrite=True)
    new_hdu.writeto(filename.replace('.fits', '.weight.fits'), overwrite=True)
    print('Masked %s pixels in %s in' % (len(mask), filename),  round(time.time()-start, 3), 'seconds.')
    return 0



class Broken_powerlaw(Fittable1DModel):
    inputs = ('x',)
    outputs = ('y',)

    compression = Parameter()
    n0 = Parameter()
    alpha1 = Parameter()
    alpha2 = Parameter()
    x_break = Parameter()

    @staticmethod
    def evaluate(x, compression, n0, alpha1, alpha2, x_break):
        alpha = np.where(x <= x_break, alpha1, alpha2)
        const = np.where(x <= x_break, compression*n0, n0)
        xx = x / x_break
        return const * xx ** (-alpha)

    @staticmethod
    def fit_deriv(x, compression, n0, alpha1, alpha2, x_break):
        alpha = np.where(x <= x_break, alpha1, alpha2)
        const = np.where(x <= x_break, compression*n0, n0)
        xx = x / x_break

        d_amplitude = xx ** (-alpha)
        d_x_break = const * alpha * d_amplitude / x_break
        d_alpha = -const * np.log(xx)
        d_alpha1 = np.where(x <= x_break, d_alpha, 0)
        d_alpha2 = np.where(x > x_break, d_alpha, 0)
        d_comp = n0 * d_amplitude
        d_n0 = d_amplitude

        return [d_comp, d_n0, d_alpha1, d_alpha2, d_x_break]

    @property
    def input_units(self):
        if self.x_break.unit is None:
            return None
        else:
            return {'x': self.x_break.unit}
    def _parameter_units_for_data_units(self, input_unit, outputs_unit):
        return OrderDict([('x_break', inputs_unit['x']), ('amplitude', outputs_unit['y'])])








